/*
 * Dump R4x00 TLB for debugging purposes.
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Ralf Baechle.
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>

#include <asm/bootinfo.h>
#include <asm/cachectl.h>
#include <asm/mipsregs.h>

static void
dump_tlb(int first, int last)
{
	int	i;
	int	wired;
	unsigned int pagemask;
	unsigned long long entryhi, entrylo0, entrylo1;

	wired = read_32bit_cp0_register(CP0_WIRED);
	printk("Wired: %d", wired);
	
	for(i=first;i<last;i++)
	{
		write_32bit_cp0_register(CP0_INDEX, i);
		__asm__ __volatile__(
			".set\tmips3\n\t"
			".set\tnoreorder\n\t"
			"nop;nop;nop;nop\n\t"
			"tlbr\n\t"
			"nop;nop;nop;nop\n\t"
			".set\treorder\n\t"
			".set\tmips0\n\t");
		pagemask = read_32bit_cp0_register(CP0_PAGEMASK);
		entryhi  = read_64bit_cp0_register(CP0_ENTRYHI);
		entrylo0 = read_64bit_cp0_register(CP0_ENTRYLO0);
		entrylo1 = read_64bit_cp0_register(CP0_ENTRYLO1);

		if((entrylo0|entrylo1) & 2)
		{
			/*
			 * Only print entries in use
			 */
			printk("\nIndex: %2d  %08x", i, pagemask);

			printk("  %08x %08x", (unsigned int)(entryhi >> 32),
			                      (unsigned int) entryhi);
			printk("  %08x %08x", (unsigned int)(entrylo0 >> 32),
			                      (unsigned int) entrylo0);
			printk("  %08x %08x", (unsigned int)(entrylo1 >> 32),
			                      (unsigned int) entrylo1);
		}
	}
	printk("\n");
}

void
dump_tlb_all(void)
{
	dump_tlb(0, boot_info.tlb_entries - 1);
}

void
dump_tlb_nonwired(void)
{
	dump_tlb(read_32bit_cp0_register(CP0_WIRED), boot_info.tlb_entries - 1);
}

#include <linux/kernel.h>
#include <linux/mm.h>

#include <asm/mipsconfig.h>
#include <asm/page.h>
#include <asm/pgtable.h>

void
dump_list_process(struct task_struct *t, void *address)
{
	pgd_t	*page_dir, *pgd;
	pmd_t	*pmd;
	pte_t	*pte, page;
	unsigned int addr;

	addr = (unsigned int) address;

	printk("Addr              == %08x\n", addr);
	printk("tasks->tss.pg_dir == %08x\n", (unsigned int) t->tss.pg_dir);
	printk("tasks->mm.pgd     == %08x\n", (unsigned int) t->mm->pgd);

	page_dir = pgd_offset(t->mm, 0);
	printk("page_dir == %08x\n", (unsigned int) page_dir);

	pgd = pgd_offset(t->mm, addr);
	printk("pgd == %08x, ", (unsigned int) pgd);

	pmd = pmd_offset(pgd, addr);
	printk("pmd == %08x, ", (unsigned int) pmd);

	pte = pte_offset(pmd, addr);
	printk("pte == %08x, ", (unsigned int) pte);

	page = *pte;
	printk("page == %08x\n", (unsigned int) pte_val(page));

}

void
dump_list_current(void *address)
{
	unsigned int addr, *pt;

	dump_list_process(current, address);

	addr = (unsigned int) address;
	pt = (unsigned int *) TLB_ROOT + (addr >> 22);
	printk("L1: *%08x == ", (unsigned int) pt);
	printk("%08x, ", *pt);

	pt = (unsigned int *) (TLBMAP + ((addr >> 10) & ~3));
	printk("L2: *%08x == ", (unsigned int) pt);
	printk("%08x\n", *pt);
}

#include <asm/segment.h>

unsigned int
vtop(void *address)
{
	pgd_t	*pgd;
	pmd_t	*pmd;
	pte_t	*pte;
	unsigned int addr, paddr;

	addr = (unsigned long) address;
	pgd = pgd_offset(current->mm, addr);
	pmd = pmd_offset(pgd, addr);
	pte = pte_offset(pmd, addr);
	paddr = (KSEG1 | (unsigned int) pte_val(*pte)) & PAGE_MASK;
	paddr |= (addr & ~PAGE_MASK);

	return paddr;
}

void
dump16(unsigned long *p)
{
	int i;

	for(i=0;i<8;i++)
	{
		printk("*%08lx == %08lx, ",
		       (unsigned long)p, (unsigned long)*p++);
		printk("*%08lx == %08lx\n",
		       (unsigned long)p, (unsigned long)*p++);
	}
}

static void
print_pte(pte_t *pte, unsigned long addr)
{
	printk("print_pte(): addr == %08lx\n", addr);
}

static void
print_active_ptes(pmd_t *pmd, unsigned long addr)
{
	int i;
	pte_t *pte;

	if (pmd_none(*pmd))
		return;
	pte = pte_offset(pmd, addr);
	for(i=0;i<PTRS_PER_PTE;i++)
	{
		print_pte(pte, addr);
		pte++;
		addr += PAGE_SIZE;
	}
}

static void
print_pmd(pmd_t *pmd, unsigned long addr)
{
	if (pmd_none(*pmd))
		return;
	if (!pmd_bad(*pmd))
		return;
	printk("print_pmd(): bad pmd for vaddr == %08lx, pmd == %08lx\n",
	       addr, pmd);
	while (1);
}

static void
print_active_pmds(pgd_t *pgd, unsigned long addr)
{
	int i;
	pmd_t *pmd;

	pmd = pmd_offset(pgd, addr);
	for(i=0;i<PTRS_PER_PMD;i++)
	{
#if 0
		print_active_ptes(pmd, addr);
#else
		print_pmd(pmd, addr);
#endif
		pmd++;
		addr += PMD_SIZE;
	}
}

void
print_active_pages(void)
{
	int i;
	struct task_struct *t = current;
	pgd_t *pgd;
	unsigned long addr = 0;

	pgd = pgd_offset(t->mm, addr);
	for(i=0;i<PTRS_PER_PGD;i++)
	{
		print_active_pmds(pgd, addr);
		pgd++;
		addr += PGDIR_SIZE;
	}
}

#define USER_PTRS_PER_PGD (TASK_SIZE/PGDIR_SIZE)

unsigned int
count_pgds(pgd_t *pgd)
{
	unsigned int i, c = 0;

	for(i=0;i<USER_PTRS_PER_PGD;i++)
	{
		if (pgd_bad(*(pgd + i)))
			c++;
	}

	return c;
}

static void
check_pmd(pmd_t *pmd, unsigned long addr, char *msg)
{
	if (pmd_none(*pmd))
		return;
	if (!pmd_bad(*pmd))
		return;
	printk("check_pmd(): bad pmd for vaddr == %08lx, pmd == %08lx\n",
	       addr, pmd);
	printk("check_pmd(): bad *pmd is %08lx.\n", (unsigned long) pmd_val(*pmd));
	printk("Stopped at %s\n", msg);

	while (1);
}

static void
check_active_pmds(pgd_t *pgd, unsigned long addr, char *msg)
{
	int i;
	pmd_t *pmd;

	if (pgd_none(*pgd))
		return;
	pmd = pmd_offset(pgd, addr);
	for(i=0;i<PTRS_PER_PMD;i++)
	{
		check_pmd(pmd, addr, msg);
		pmd++;
		addr += PMD_SIZE;
	}
}

void
check_active_pages(char *msg)
{
	int i;
	struct task_struct *t = current;
	pgd_t *pgd;
	unsigned long addr = 0;
	int c;

	if((c = count_pgds((pgd_t *)0xa0001000)))
	{
		printk("Kernel Page directory at %08lx is corrupted.\n", msg);
		printk("It contains %d dirty pages.\n", c);
		printk("Stopped at %s\n", msg);
	}

	pgd = pgd_offset(t->mm, addr);
	for(i=0;i<PTRS_PER_PGD;i++)
	{
		check_active_pmds(pgd, addr, msg);
		pgd++;
		addr += PGDIR_SIZE;
	}
}
